@@ -1,6 +1,6 @@
 /*  Plzip - A parallel compressor compatible with lzip
     Copyright (C) 2009 Laszlo Ersek.
-    Copyright (C) 2009, 2010 Antonio Diaz Diaz.
+    Copyright (C) 2009, 2010, 2011, 2012, 2013 Antonio Diaz Diaz.
 
     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -16,7 +16,7 @@
     along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
 /*
-    Return values: 0 for a normal exit, 1 for environmental problems
+    Exit status: 0 for a normal exit, 1 for environmental problems
     (file not found, invalid flags, I/O errors, etc), 2 to indicate a
     corrupt or invalid input file, 3 for an internal consistency error
     (eg, bug) which caused plzip to panic.
@@ -34,37 +34,42 @@
 #include <string>
 #include <vector>
 #include <fcntl.h>
-#include <inttypes.h>
 #include <pthread.h>
+#include <stdint.h>
 #include <unistd.h>
 #include <utime.h>
 #include <sys/stat.h>
 #include <lzlib.h>
+#if defined(__MSVCRT__)
+#include <io.h>
+#define fchmod(x,y) 0
+#define fchown(x,y,z) 0
+#define strtoull std::strtoul
+#define SIGHUP SIGTERM
+#define S_ISSOCK(x) 0
+#define S_IRGRP 0
+#define S_IWGRP 0
+#define S_IROTH 0
+#define S_IWOTH 0
+#endif
+#if defined(__OS2__)
+#include <io.h>
+#endif
 
 #include "arg_parser.h"
-#include "plzip.h"
+#include "lzip.h"
 
 #if CHAR_BIT != 8
 #error "Environments where CHAR_BIT != 8 are not supported."
 #endif
 
-#ifndef LLONG_MAX
-#define LLONG_MAX  0x7FFFFFFFFFFFFFFFLL
-#endif
-#ifndef LLONG_MIN
-#define LLONG_MIN  (-LLONG_MAX - 1LL)
-#endif
-#ifndef ULLONG_MAX
-#define ULLONG_MAX 0xFFFFFFFFFFFFFFFFULL
-#endif
-
 
 namespace {
 
+const char * const Program_name = "Plzip";
+const char * const program_name = "plzip";
+const char * const program_year = "2013";
 const char * invocation_name = 0;
-const char * const Program_name    = "Plzip";
-const char * const program_name    = "plzip";
-const char * const program_year    = "2010";
 
 #ifdef O_BINARY
 const int o_binary = O_BINARY;
@@ -83,97 +88,84 @@ struct Lzma_options
   int match_len_limit;		// 5..273
   };
 
-enum Mode { m_compress = 0, m_decompress, m_test };
+enum Mode { m_compress, m_decompress, m_test };
 
 std::string output_filename;
 int outfd = -1;
-mode_t outfd_mode = S_IRUSR | S_IWUSR;
+const mode_t usr_rw = S_IRUSR | S_IWUSR;
+const mode_t all_rw = usr_rw | S_IRGRP | S_IWGRP | S_IROTH | S_IWOTH;
+mode_t outfd_mode = usr_rw;
 bool delete_output_on_interrupt = false;
 pthread_t main_thread;
 pid_t main_thread_pid;
 
 
-void show_help() throw()
+void show_help( const long num_online )
   {
   std::printf( "%s - A parallel compressor compatible with lzip.\n", Program_name );
   std::printf( "\nUsage: %s [options] [files]\n", invocation_name );
-  std::printf( "\nOptions:\n" );
-  std::printf( "  -h, --help                 display this help and exit\n" );
-  std::printf( "  -V, --version              output version information and exit\n" );
-  std::printf( "  -B, --data-size=<n>        set input data block size in bytes\n" );
-  std::printf( "  -c, --stdout               send output to standard output\n" );
-  std::printf( "  -d, --decompress           decompress\n" );
-  std::printf( "  -f, --force                overwrite existing output files\n" );
-  std::printf( "  -k, --keep                 keep (don't delete) input files\n" );
-  std::printf( "  -m, --match-length=<n>     set match length limit in bytes [80]\n" );
-  std::printf( "  -n, --threads=<n>          set the number of (de)compression threads\n" );
-  std::printf( "  -o, --output=<file>        if reading stdin, place the output into <file>\n" );
-  std::printf( "  -q, --quiet                suppress all messages\n" );
-  std::printf( "  -s, --dictionary-size=<n>  set dictionary size limit in bytes [8MiB]\n" );
-  std::printf( "  -t, --test                 test compressed file integrity\n" );
-  std::printf( "  -v, --verbose              be verbose (a 2nd -v gives more)\n" );
-  std::printf( "  -1 .. -9                   set compression level [default 6]\n" );
-  std::printf( "      --fast                 alias for -1\n" );
-  std::printf( "      --best                 alias for -9\n" );
+  std::printf( "\nOptions:\n"
+               "  -h, --help                     display this help and exit\n"
+               "  -V, --version                  output version information and exit\n"
+               "  -B, --data-size=<bytes>        set input data block size in bytes\n"
+               "  -c, --stdout                   send output to standard output\n"
+               "  -d, --decompress               decompress\n"
+               "  -f, --force                    overwrite existing output files\n"
+               "  -F, --recompress               force recompression of compressed files\n"
+               "  -k, --keep                     keep (don't delete) input files\n"
+               "  -m, --match-length=<bytes>     set match length limit in bytes [36]\n"
+               "  -n, --threads=<n>              set number of (de)compression threads [%ld]\n"
+               "  -o, --output=<file>            if reading stdin, place the output into <file>\n"
+               "  -q, --quiet                    suppress all messages\n"
+               "  -s, --dictionary-size=<bytes>  set dictionary size limit in bytes [8MiB]\n"
+               "  -t, --test                     test compressed file integrity\n"
+               "  -v, --verbose                  be verbose (a 2nd -v gives more)\n"
+               "  -1 .. -9                       set compression level [default 6]\n"
+               "      --fast                     alias for -1\n"
+               "      --best                     alias for -9\n", num_online );
   if( verbosity > 0 )
     {
     std::printf( "  -D, --debug=<level>        (0-1) print debug statistics to stderr\n" );
     }
-  std::printf( "If no file names are given, %s compresses or decompresses\n", program_name );
-  std::printf( "from standard input to standard output.\n" );
-  std::printf( "Numbers may be followed by a multiplier: k = kB = 10^3 = 1000,\n" );
-  std::printf( "Ki = KiB = 2^10 = 1024, M = 10^6, Mi = 2^20, G = 10^9, Gi = 2^30, etc...\n" );
-  std::printf( "\nReport bugs to lzip-bug@nongnu.org\n" );
-  std::printf( "Plzip home page: http://www.nongnu.org/lzip/plzip.html\n" );
+  std::printf( "If no file names are given, plzip compresses or decompresses\n"
+               "from standard input to standard output.\n"
+               "Numbers may be followed by a multiplier: k = kB = 10^3 = 1000,\n"
+               "Ki = KiB = 2^10 = 1024, M = 10^6, Mi = 2^20, G = 10^9, Gi = 2^30, etc...\n"
+               "The bidimensional parameter space of LZMA can't be mapped to a linear\n"
+               "scale optimal for all files. If your files are large, very repetitive,\n"
+               "etc, you may need to use the --match-length and --dictionary-size\n"
+               "options directly to achieve optimal performance.\n"
+               "\nExit status: 0 for a normal exit, 1 for environmental problems (file\n"
+               "not found, invalid flags, I/O errors, etc), 2 to indicate a corrupt or\n"
+               "invalid input file, 3 for an internal consistency error (eg, bug) which\n"
+               "caused plzip to panic.\n"
+               "\nReport bugs to lzip-bug@nongnu.org\n"
+               "Plzip home page: http://www.nongnu.org/lzip/plzip.html\n" );
   }
 
 
-void show_version() throw()
+void show_version()
   {
   std::printf( "%s %s\n", Program_name, PROGVERSION );
-  std::printf( "Copyright (C) 2009 Laszlo Ersek.\n" );
-  std::printf( "Copyright (C) %s Antonio Diaz Diaz.\n", program_year );
+  std::printf( "Copyright (C) 2009 Laszlo Ersek.\n"
+               "Copyright (C) %s Antonio Diaz Diaz.\n", program_year );
   std::printf( "Using Lzlib %s\n", LZ_version() );
-  std::printf( "License GPLv3+: GNU GPL version 3 or later <http://gnu.org/licenses/gpl.html>\n" );
-  std::printf( "This is free software: you are free to change and redistribute it.\n" );
-  std::printf( "There is NO WARRANTY, to the extent permitted by law.\n" );
+  std::printf( "License GPLv3+: GNU GPL version 3 or later <http://gnu.org/licenses/gpl.html>\n"
+               "This is free software: you are free to change and redistribute it.\n"
+               "There is NO WARRANTY, to the extent permitted by law.\n" );
   }
 
 
-const char * format_num( long long num, long long limit = 9999,
-                         const int set_prefix = 0 ) throw()
-  {
-  const char * const si_prefix[8] =
-    { "k", "M", "G", "T", "P", "E", "Z", "Y" };
-  const char * const binary_prefix[8] =
-    { "Ki", "Mi", "Gi", "Ti", "Pi", "Ei", "Zi", "Yi" };
-  static bool si = false;
-  static char buf[16];
-
-  if( set_prefix ) si = ( set_prefix > 0 );
-  const int factor = ( si ) ? 1000 : 1024;
-  const char * const *prefix = ( si ) ? si_prefix : binary_prefix;
-  const char *p = "";
-  limit = std::max( 999LL, std::min( 999999LL, limit ) );
-
-  for( int i = 0; i < 8 && ( llabs( num ) > limit ||
-       ( llabs( num ) >= factor && num % factor == 0 ) ); ++i )
-    { num /= factor; p = prefix[i]; }
-  snprintf( buf, sizeof buf, "%lld %s", num, p );
-  return buf;
-  }
-
-
-long long getnum( const char * const ptr, const int bs = 0,
-                  const long long llimit = LLONG_MIN + 1,
-                  const long long ulimit = LLONG_MAX ) throw()
+unsigned long long getnum( const char * const ptr,
+                           const unsigned long long llimit,
+                           const unsigned long long ulimit )
   {
+  char * tail;
   errno = 0;
-  char *tail;
-  long long result = strtoll( ptr, &tail, 0 );
+  unsigned long long result = strtoull( ptr, &tail, 0 );
   if( tail == ptr )
     {
-    show_error( "bad or missing numerical argument", 0, true );
+    show_error( "Bad or missing numerical argument.", 0, true );
     std::exit( 1 );
     }
 
@@ -185,9 +177,6 @@ long long getnum( const char * const ptr, const int bs = 0,
     switch( tail[0] )
       {
       case ' ': break;
-      case 'b': if( bs > 0 ) { factor = bs; exponent = 1; }
-                else bad_multiplier = true;
-                break;
       case 'Y': exponent = 8; break;
       case 'Z': exponent = 7; break;
       case 'E': exponent = 6; break;
@@ -203,37 +192,37 @@ long long getnum( const char * const ptr, const int bs = 0,
       }
     if( bad_multiplier )
       {
-      show_error( "bad multiplier in numerical argument", 0, true );
+      show_error( "Bad multiplier in numerical argument.", 0, true );
       std::exit( 1 );
       }
     for( int i = 0; i < exponent; ++i )
       {
-      if( LLONG_MAX / factor >= llabs( result ) ) result *= factor;
+      if( ulimit / factor >= result ) result *= factor;
       else { errno = ERANGE; break; }
       }
     }
   if( !errno && ( result < llimit || result > ulimit ) ) errno = ERANGE;
   if( errno )
     {
-    show_error( "numerical argument out of limits" );
+    show_error( "Numerical argument out of limits." );
     std::exit( 1 );
     }
   return result;
   }
 
 
-int get_dict_size( const char * const arg ) throw()
+int get_dict_size( const char * const arg )
   {
-  char *tail;
+  char * tail;
   int bits = std::strtol( arg, &tail, 0 );
   if( bits >= LZ_min_dictionary_bits() &&
       bits <= LZ_max_dictionary_bits() && *tail == 0 )
     return ( 1 << bits );
-  return getnum( arg, 0, LZ_min_dictionary_size(), LZ_max_dictionary_size() );
+  return getnum( arg, LZ_min_dictionary_size(), LZ_max_dictionary_size() );
   }
 
 
-int extension_index( const std::string & name ) throw()
+int extension_index( const std::string & name )
   {
   for( int i = 0; known_extensions[i].from; ++i )
     {
@@ -246,39 +235,40 @@ int extension_index( const std::string & name ) throw()
   }
 
 
-int open_instream( const std::string & name, struct stat * const in_statsp,
+int open_instream( const char * const name, struct stat * const in_statsp,
                    const Mode program_mode, const int eindex,
-                   const bool force, const bool to_stdout ) throw()
+                   const bool recompress, const bool to_stdout )
   {
   int infd = -1;
-  if( program_mode == m_compress && !force && eindex >= 0 )
+  if( program_mode == m_compress && !recompress && eindex >= 0 )
     {
     if( verbosity >= 0 )
-      std::fprintf( stderr, "%s: input file `%s' already has `%s' suffix.\n",
-                    program_name, name.c_str(),
-                    known_extensions[eindex].from );
+      std::fprintf( stderr, "%s: Input file '%s' already has '%s' suffix.\n",
+                    program_name, name, known_extensions[eindex].from );
     }
   else
     {
-    infd = open( name.c_str(), O_RDONLY | o_binary );
+    infd = open( name, O_RDONLY | o_binary );
     if( infd < 0 )
       {
       if( verbosity >= 0 )
-        std::fprintf( stderr, "%s: Can't open input file `%s': %s.\n",
-                      program_name, name.c_str(), std::strerror( errno ) );
+        std::fprintf( stderr, "%s: Can't open input file '%s': %s.\n",
+                      program_name, name, std::strerror( errno ) );
       }
     else
       {
       const int i = fstat( infd, in_statsp );
-      const mode_t & mode = in_statsp->st_mode;
-      if( i < 0 || !( S_ISREG( mode ) || ( to_stdout &&
-                      ( S_ISFIFO( mode ) || S_ISSOCK( mode ) ||
-                        S_ISBLK( mode ) || S_ISCHR( mode ) ) ) ) )
+      const mode_t mode = in_statsp->st_mode;
+      const bool can_read = ( i == 0 &&
+                              ( S_ISBLK( mode ) || S_ISCHR( mode ) ||
+                                S_ISFIFO( mode ) || S_ISSOCK( mode ) ) );
+      if( i != 0 || ( !S_ISREG( mode ) && ( !to_stdout || !can_read ) ) )
         {
         if( verbosity >= 0 )
-          std::fprintf( stderr, "%s: input file `%s' is not a regular file%s.\n",
-                        program_name, name.c_str(),
-                        to_stdout ? "" : " and `--stdout' was not specified" );
+          std::fprintf( stderr, "%s: Input file '%s' is not a regular file%s.\n",
+                        program_name, name,
+                        ( can_read && !to_stdout ) ?
+                        " and '--stdout' was not specified" : "" );
         close( infd );
         infd = -1;
         }
@@ -288,14 +278,14 @@ int open_instream( const std::string & name, struct stat * const in_statsp,
   }
 
 
-void set_c_outname( const std::string & name ) throw()
+void set_c_outname( const std::string & name )
   {
   output_filename = name;
   output_filename += known_extensions[0].from;
   }
 
 
-void set_d_outname( const std::string & name, const int i ) throw()
+void set_d_outname( const std::string & name, const int i )
   {
   if( i >= 0 )
     {
@@ -308,38 +298,34 @@ void set_d_outname( const std::string & name, const int i ) throw()
       }
     }
   output_filename = name; output_filename += ".out";
-  if( verbosity >= 0 )
-    std::fprintf( stderr, "%s: can't guess original name for `%s' -- using `%s'.\n",
+  if( verbosity >= 1 )
+    std::fprintf( stderr, "%s: Can't guess original name for '%s' -- using '%s'.\n",
                   program_name, name.c_str(), output_filename.c_str() );
   }
 
 
-bool open_outstream( const bool force ) throw()
+bool open_outstream( const bool force )
   {
   int flags = O_CREAT | O_WRONLY | o_binary;
   if( force ) flags |= O_TRUNC; else flags |= O_EXCL;
 
   outfd = open( output_filename.c_str(), flags, outfd_mode );
-  if( outfd < 0 )
+  if( outfd < 0 && verbosity >= 0 )
     {
-    if( errno == EEXIST ) outfd = -2; else outfd = -1;
-    if( verbosity >= 0 )
-      {
-      if( outfd == -2 )
-        std::fprintf( stderr, "%s: Output file %s already exists, skipping.\n",
-                      program_name, output_filename.c_str() );
-      else
-        std::fprintf( stderr, "%s: Can't create output file `%s': %s.\n",
-                      program_name, output_filename.c_str(), std::strerror( errno ) );
-      }
+    if( errno == EEXIST )
+      std::fprintf( stderr, "%s: Output file '%s' already exists, skipping.\n",
+                    program_name, output_filename.c_str() );
+    else
+      std::fprintf( stderr, "%s: Can't create output file '%s': %s.\n",
+                    program_name, output_filename.c_str(), std::strerror( errno ) );
     }
   return ( outfd >= 0 );
   }
 
 
-bool check_tty( const int infd, const Mode program_mode ) throw()
+bool check_tty( const int infd, const Mode program_mode )
   {
-  if( program_mode == m_compress && isatty( outfd ) )
+  if( program_mode == m_compress && outfd >= 0 && isatty( outfd ) )
     {
     show_error( "I won't write compressed data to a terminal.", 0, true );
     return false;
@@ -354,16 +340,16 @@ bool check_tty( const int infd, const Mode program_mode ) throw()
   }
 
 
-void cleanup_and_fail( const int retval ) throw()
+void cleanup_and_fail( const int retval )
   {
   if( delete_output_on_interrupt )
     {
     delete_output_on_interrupt = false;
     if( verbosity >= 0 )
-      std::fprintf( stderr, "%s: Deleting output file `%s', if it exists.\n",
-               program_name, output_filename.c_str() );
+      std::fprintf( stderr, "%s: Deleting output file '%s', if it exists.\n",
+                    program_name, output_filename.c_str() );
     if( outfd >= 0 ) { close( outfd ); outfd = -1; }
-    if( std::remove( output_filename.c_str() ) != 0 )
+    if( std::remove( output_filename.c_str() ) != 0 && errno != ENOENT )
       show_error( "WARNING: deletion of output file (apparently) failed." );
     }
   std::exit( retval );
@@ -373,51 +359,44 @@ void cleanup_and_fail( const int retval ) throw()
      // Set permissions, owner and times.
 void close_and_set_permissions( const struct stat * const in_statsp )
   {
-  bool error = false;
+  bool warning = false;
   if( in_statsp )
     {
-    if( fchmod( outfd, in_statsp->st_mode ) != 0 ) error = true;
-    else (void)fchown( outfd, in_statsp->st_uid, in_statsp->st_gid );
     // fchown will in many cases return with EPERM, which can be safely ignored.
+    if( ( fchown( outfd, in_statsp->st_uid, in_statsp->st_gid ) != 0 &&
+          errno != EPERM ) ||
+        fchmod( outfd, in_statsp->st_mode ) != 0 ) warning = true;
     }
-  if( close( outfd ) == 0 ) outfd = -1;
-  else cleanup_and_fail( 1 );
+  if( close( outfd ) != 0 ) cleanup_and_fail( 1 );
+  outfd = -1;
   delete_output_on_interrupt = false;
-  if( !in_statsp ) return;
-  if( !error )
+  if( in_statsp )
     {
     struct utimbuf t;
     t.actime = in_statsp->st_atime;
     t.modtime = in_statsp->st_mtime;
-    if( utime( output_filename.c_str(), &t ) != 0 ) error = true;
-    }
-  if( error )
-    {
-    show_error( "I can't change output file attributes." );
-    cleanup_and_fail( 1 );
+    if( utime( output_filename.c_str(), &t ) != 0 ) warning = true;
     }
+  if( warning && verbosity >= 1 )
+    show_error( "Can't change output file attributes." );
   }
 
 
-extern "C" void signal_handler( int sig ) throw()
+extern "C" void signal_handler( int sig )
   {
   if( !pthread_equal( pthread_self(), main_thread ) )
     kill( main_thread_pid, sig );
-  if( sig != SIGUSR1 )
+  if( sig != SIGUSR1 && sig != SIGUSR2 )
     show_error( "Control-C or similar caught, quitting." );
-  cleanup_and_fail( 1 );
+  cleanup_and_fail( ( sig != SIGUSR2 ) ? 1 : 2 );
   }
 
 
-void set_signals( const bool to_file ) throw()
+void set_signals()
   {
-  if( to_file )
-    {
-    std::signal( SIGHUP, signal_handler );
-    std::signal( SIGINT, signal_handler );
-    std::signal( SIGTERM, signal_handler );
-    }
-  std::signal( SIGUSR1, signal_handler );
+  std::signal( SIGHUP, signal_handler );
+  std::signal( SIGINT, signal_handler );
+  std::signal( SIGTERM, signal_handler );
   }
 
 } // end namespace
@@ -426,13 +405,15 @@ void set_signals( const bool to_file ) throw()
 int verbosity = 0;
 
 
-// This can be called from any thread, main thread or sub-threads alike, since
-// they all call common helper functions that call fatal() in case of an error.
+// This can be called from any thread, main thread or sub-threads alike,
+// since they all call common helper functions that call fatal() in case
+// of an error.
 //
-void fatal() { signal_handler( SIGUSR1 ); }
+void fatal( const int retval )
+  { signal_handler( ( retval != 2 ) ? SIGUSR1 : SIGUSR2 ); }
 
 
-void Pretty_print::operator()( const char * const msg ) const throw()
+void Pretty_print::operator()( const char * const msg ) const
   {
   if( verbosity >= 0 )
     {
@@ -440,7 +421,7 @@ void Pretty_print::operator()( const char * const msg ) const throw()
       {
       first_post = false;
       std::fprintf( stderr, "  %s: ", name_.c_str() );
-      for( unsigned int i = 0; i < longest_name - name_.size(); ++i )
+      for( unsigned i = 0; i < longest_name - name_.size(); ++i )
         std::fprintf( stderr, " " );
       if( !msg ) std::fflush( stderr );
       }
@@ -449,95 +430,61 @@ void Pretty_print::operator()( const char * const msg ) const throw()
   }
 
 
-void show_error( const char * const msg, const int errcode, const bool help ) throw()
+void show_error( const char * const msg, const int errcode, const bool help )
   {
   if( verbosity >= 0 )
     {
-    if( msg && msg[0] != 0 )
+    if( msg && msg[0] )
       {
       std::fprintf( stderr, "%s: %s", program_name, msg );
-      if( errcode > 0 ) std::fprintf( stderr, ": %s", std::strerror( errcode ) );
+      if( errcode > 0 )
+        std::fprintf( stderr, ": %s", std::strerror( errcode ) );
       std::fprintf( stderr, "\n" );
       }
-    if( help && invocation_name && invocation_name[0] != 0 )
-      std::fprintf( stderr, "Try `%s --help' for more information.\n", invocation_name );
+    if( help )
+      std::fprintf( stderr, "Try '%s --help' for more information.\n",
+                    invocation_name );
     }
   }
 
 
 void internal_error( const char * const msg )
   {
-  std::string s( "internal error: " ); s += msg;
-  show_error( s.c_str() );
+  if( verbosity >= 0 )
+    std::fprintf( stderr, "%s: internal error: %s.\n", program_name, msg );
   std::exit( 3 );
   }
 
 
-// Returns the number of bytes really read.
-// If (returned value < size) and (errno == 0), means EOF was reached.
-//
-int readblock( const int fd, uint8_t * const buf, const int size ) throw()
-  {
-  int rest = size;
-  errno = 0;
-  while( rest > 0 )
-    {
-    errno = 0;
-    const int n = read( fd, buf + size - rest, rest );
-    if( n > 0 ) rest -= n;
-    else if( n == 0 ) break;
-    else if( errno != EINTR && errno != EAGAIN ) break;
-    }
-  return ( rest > 0 ) ? size - rest : size;
-  }
-
-
-// Returns the number of bytes really written.
-// If (returned value < size), it is always an error.
-//
-int writeblock( const int fd, const uint8_t * const buf, const int size ) throw()
-  {
-  int rest = size;
-  errno = 0;
-  while( rest > 0 )
-    {
-    errno = 0;
-    const int n = write( fd, buf + size - rest, rest );
-    if( n > 0 ) rest -= n;
-    else if( errno && errno != EINTR && errno != EAGAIN ) break;
-    }
-  return ( rest > 0 ) ? size - rest : size;
-  }
-
-
 int main( const int argc, const char * const argv[] )
   {
   // Mapping from gzip/bzip2 style 1..9 compression modes
   // to the corresponding LZMA compression modes.
   const Lzma_options option_mapping[] =
     {
-    { 1 << 16,   5 },		// -0
-    { 1 << 20,  10 },		// -1
-    { 3 << 19,  12 },		// -2
-    { 1 << 21,  17 },		// -3
-    { 3 << 20,  26 },		// -4
-    { 1 << 22,  44 },		// -5
-    { 1 << 23,  80 },		// -6
-    { 1 << 24, 108 },		// -7
-    { 3 << 23, 163 },		// -8
+    { 1 << 20,   5 },		// -0
+    { 1 << 20,   5 },		// -1
+    { 3 << 19,   6 },		// -2
+    { 1 << 21,   8 },		// -3
+    { 3 << 20,  12 },		// -4
+    { 1 << 22,  20 },		// -5
+    { 1 << 23,  36 },		// -6
+    { 1 << 24,  68 },		// -7
+    { 3 << 23, 132 },		// -8
     { 1 << 25, 273 } };		// -9
   Lzma_options encoder_options = option_mapping[6];	// default = "-6"
-  int data_size = 0;
-  int debug_level = 0;
-  int infd = -1;
-  int num_workers = 0;		// Start this many worker threads
-  Mode program_mode = m_compress;
-  bool force = false;
-  bool keep_input_files = false;
-  bool to_stdout = false;
   std::string input_filename;
   std::string default_output_filename;
   std::vector< std::string > filenames;
+  int data_size = 0;
+  int debug_level = 0;
+  int infd = -1;
+  int num_workers = 0;		// start this many worker threads
+  Mode program_mode = m_compress;
+  bool force = false;
+  bool keep_input_files = false;
+  bool recompress = false;
+  bool to_stdout = false;
   invocation_name = argv[0];
   main_thread = pthread_self();
   main_thread_pid = getpid();
@@ -545,17 +492,15 @@ int main( const int argc, const char * const argv[] )
   if( LZ_version()[0] != LZ_version_string[0] )
     internal_error( "bad library version" );
 
-  const int slots_per_worker = 2;
+  const long num_online = std::max( 1L, sysconf( _SC_NPROCESSORS_ONLN ) );
   long max_workers = sysconf( _SC_THREAD_THREADS_MAX );
-  if( max_workers < 1 || max_workers > INT_MAX / slots_per_worker )
-    max_workers = INT_MAX / slots_per_worker;
-  if( max_workers > INT_MAX / (int)sizeof (pthread_t) )
+  if( max_workers < 1 || max_workers > INT_MAX / (int)sizeof (pthread_t) )
     max_workers = INT_MAX / sizeof (pthread_t);
 
   const Arg_parser::Option options[] =
     {
-    { '0',  0,                Arg_parser::no  },
-    { '1', "fast",            Arg_parser::no  },
+    { '0', "fast",            Arg_parser::no  },
+    { '1',  0,                Arg_parser::no  },
     { '2',  0,                Arg_parser::no  },
     { '3',  0,                Arg_parser::no  },
     { '4',  0,                Arg_parser::no  },
@@ -568,9 +513,9 @@ int main( const int argc, const char * const argv[] )
     { 'B', "data-size",       Arg_parser::yes },
     { 'c', "stdout",          Arg_parser::no  },
     { 'd', "decompress",      Arg_parser::no  },
-    { 'e', "extreme",         Arg_parser::no  },
     { 'D', "debug",           Arg_parser::yes },
     { 'f', "force",           Arg_parser::no  },
+    { 'F', "recompress",      Arg_parser::no  },
     { 'h', "help",            Arg_parser::no  },
     { 'k', "keep",            Arg_parser::no  },
     { 'm', "match-length",    Arg_parser::yes },
@@ -584,7 +529,7 @@ int main( const int argc, const char * const argv[] )
     { 'V', "version",         Arg_parser::no  },
     {  0 ,  0,                Arg_parser::no  } };
 
-  Arg_parser parser( argc, argv, options );
+  const Arg_parser parser( argc, argv, options );
   if( parser.error().size() )				// bad option
     { show_error( parser.error().c_str(), 0, true ); return 1; }
 
@@ -596,24 +541,25 @@ int main( const int argc, const char * const argv[] )
     const char * const arg = parser.argument( argind ).c_str();
     switch( code )
       {
-      case '0': case '1': case '2': case '3': case '4':
+      case '0':
+      case '1': case '2': case '3': case '4':
       case '5': case '6': case '7': case '8': case '9':
                 encoder_options = option_mapping[code-'0']; break;
       case 'b': break;
-      case 'B': data_size = getnum( arg, 0, 2 * LZ_min_dictionary_size(),
+      case 'B': data_size = getnum( arg, 2 * LZ_min_dictionary_size(),
                                     2 * LZ_max_dictionary_size() ); break;
       case 'c': to_stdout = true; break;
       case 'd': program_mode = m_decompress; break;
-      case 'D': debug_level = getnum( arg, 0, 0, 3 ); break;
-      case 'e': break;
+      case 'D': debug_level = getnum( arg, 0, 3 ); break;
       case 'f': force = true; break;
-      case 'h': show_help(); return 0;
+      case 'F': recompress = true; break;
+      case 'h': show_help( num_online ); return 0;
       case 'k': keep_input_files = true; break;
       case 'm': encoder_options.match_len_limit =
-                  getnum( arg, 0, LZ_min_match_len_limit(),
-                                  LZ_max_match_len_limit() ); break;
+                  getnum( arg, LZ_min_match_len_limit(),
+                               LZ_max_match_len_limit() ); break;
+      case 'n': num_workers = getnum( arg, 1, max_workers ); break;
       case 'o': default_output_filename = arg; break;
-      case 'n': num_workers = getnum( arg, 0, 1, max_workers ); break;
       case 'q': verbosity = -1; break;
       case 's': encoder_options.dictionary_size = get_dict_size( arg );
                 break;
@@ -623,7 +569,15 @@ int main( const int argc, const char * const argv[] )
       case 'V': show_version(); return 0;
       default : internal_error( "uncaught option" );
       }
-    }
+    } // end process options
+
+#if defined(__MSVCRT__) || defined(__OS2__)
+  setmode( STDIN_FILENO, O_BINARY );
+  setmode( STDOUT_FILENO, O_BINARY );
+#endif
+
+  if( program_mode == m_test )
+    outfd = -1;
 
   if( data_size <= 0 )
     data_size = 2 * std::max( 65536, encoder_options.dictionary_size );
@@ -631,30 +585,26 @@ int main( const int argc, const char * const argv[] )
     encoder_options.dictionary_size = std::max( data_size, LZ_min_dictionary_size() );
 
   if( num_workers <= 0 )
-    {
-    long num_online = sysconf( _SC_NPROCESSORS_ONLN );
-    if( num_online <= 0 ) num_online = 1;
     num_workers = std::min( num_online, max_workers );
-    }
-  const int num_slots = num_workers * slots_per_worker;
 
   bool filenames_given = false;
   for( ; argind < parser.arguments(); ++argind )
     {
-    if( parser.argument( argind ) != "-" ) filenames_given = true;
     filenames.push_back( parser.argument( argind ) );
+    if( filenames.back() != "-" ) filenames_given = true;
     }
 
   if( filenames.empty() ) filenames.push_back("-");
-  set_signals( !to_stdout && program_mode != m_test &&
-               ( filenames_given || default_output_filename.size() ) );
+  if( !to_stdout && program_mode != m_test &&
+      ( filenames_given || default_output_filename.size() ) )
+    set_signals();
+  std::signal( SIGUSR1, signal_handler );
+  std::signal( SIGUSR2, signal_handler );
 
   Pretty_print pp( filenames );
-  if( program_mode == m_test )
-    outfd = -1;
 
   int retval = 0;
-  for( unsigned int i = 0; i < filenames.size(); ++i )
+  for( unsigned i = 0; i < filenames.size(); ++i )
     {
     struct stat in_stats;
     output_filename.clear();
@@ -672,10 +622,10 @@ int main( const int argc, const char * const argv[] )
           if( program_mode == m_compress )
             set_c_outname( default_output_filename );
           else output_filename = default_output_filename;
-          outfd_mode = S_IRUSR | S_IWUSR | S_IRGRP | S_IWGRP | S_IROTH | S_IWOTH;
+          outfd_mode = all_rw;
           if( !open_outstream( force ) )
             {
-            if( outfd == -1 && retval < 1 ) retval = 1;
+            if( retval < 1 ) retval = 1;
             close( infd ); infd = -1;
             continue;
             }
@@ -686,8 +636,8 @@ int main( const int argc, const char * const argv[] )
       {
       input_filename = filenames[i];
       const int eindex = extension_index( input_filename );
-      infd = open_instream( input_filename, &in_stats, program_mode,
-                            eindex, force, to_stdout );
+      infd = open_instream( input_filename.c_str(), &in_stats, program_mode,
+                            eindex, recompress, to_stdout );
       if( infd < 0 ) { if( retval < 1 ) retval = 1; continue; }
       if( program_mode != m_test )
         {
@@ -697,10 +647,10 @@ int main( const int argc, const char * const argv[] )
           if( program_mode == m_compress )
             set_c_outname( input_filename );
           else set_d_outname( input_filename, eindex );
-          outfd_mode = S_IRUSR | S_IWUSR;
+          outfd_mode = usr_rw;
           if( !open_outstream( force ) )
             {
-            if( outfd == -1 && retval < 1 ) retval = 1;
+            if( retval < 1 ) retval = 1;
             close( infd ); infd = -1;
             continue;
             }
@@ -713,16 +663,17 @@ int main( const int argc, const char * const argv[] )
     if( output_filename.size() && !to_stdout && program_mode != m_test )
       delete_output_on_interrupt = true;
     const struct stat * const in_statsp = input_filename.size() ? &in_stats : 0;
+    const bool infd_isreg = in_statsp && S_ISREG( in_statsp->st_mode );
     pp.set_name( input_filename );
     if( verbosity >= 1 ) pp();
-    int tmp = 0;
+    int tmp;
     if( program_mode == m_compress )
       tmp = compress( data_size, encoder_options.dictionary_size,
-                      encoder_options.match_len_limit, num_workers,
-                      num_slots, infd, outfd, pp, debug_level );
+                      encoder_options.match_len_limit,
+                      num_workers, infd, outfd, pp, debug_level );
     else
-      tmp = decompress( num_workers, num_slots, infd, outfd, pp,
-                        debug_level, program_mode == m_test );
+      tmp = decompress( num_workers, infd, outfd, pp, debug_level,
+                        program_mode == m_test, infd_isreg );
     if( tmp > retval ) retval = tmp;
     if( tmp && program_mode != m_test ) cleanup_and_fail( retval );
 
@@ -737,9 +688,7 @@ int main( const int argc, const char * const argv[] )
     }
   if( outfd >= 0 && close( outfd ) != 0 )
     {
-    if( verbosity >= 0 )
-      std::fprintf( stderr, "%s: Can't close stdout: %s.\n",
-                    program_name, std::strerror( errno ) );
+    show_error( "Can't close stdout", errno );
     if( retval < 1 ) retval = 1;
     }
   return retval;
